# Copyright 2019-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import sys
import types
import unittest

import yaml

from charmhelpers.core import unitdata

from fixtures import Fixture, MonkeyPatch, EnvironmentVariable, MockPatch
from systemfixtures import FakeFilesystem, FakeProcesses, FakeGroups, FakeUsers


CHARM_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../'))


class BaseTestCase(unittest.TestCase):
    def useFixture(self, fixture):
        fixture.setUp()
        self.addCleanup(fixture.cleanUp)
        return fixture


class ConfigOverrideFixture(Fixture):
    def __init__(self, new_config={}):
        super().__init__()
        self._new_config = self._read_config()
        self.update(new_config)

    def update(self, new_config):
        self._new_config.update(new_config)

    def _read_config(self):
        # read config.yaml
        default_config = {}
        with open(os.path.join(CHARM_DIR, 'config.yaml')) as fd:
            config_yaml = fd.read()
            for k, v in yaml.safe_load(config_yaml)['options'].items():
                default_config[k] = v['default']
        return default_config

    def _setUp(self):
        self.useFixture(
            MonkeyPatch(
                'charmhelpers.core.hookenv.config', lambda: self._new_config.copy()
            )
        )


class LayerConfigOverrideFixture(Fixture):
    def __init__(self, new_config={}):
        super().__init__()
        self._new_config = self._read_config()
        self.update(new_config)

    def update(self, new_config):
        for layer_name, new_layer_config in new_config.items():
            self._new_config.setdefault(layer_name, {}).update(new_layer_config)

    def _read_config(self):
        # read layer.yaml
        with open(os.path.join(CHARM_DIR, 'layer.yaml')) as fd:
            layer_yaml = fd.read()
            return yaml.safe_load(layer_yaml)['options']

    def _setUp(self):
        def options(section=None):
            if section and section in self._new_config:
                return self._new_config.get(section).copy()
            else:
                return self._new_config.copy()

        self.useFixture(MonkeyPatch('charms.layer.options', options))


class UnitDataFixture(Fixture):
    """override unitdata._KV with a memory version"""

    def __init__(self):
        super().__init__()
        self.unitdata = unitdata.Storage(':memory:')

    def _setUp(self):
        unitdata._KV = self.unitdata
        self.addCleanup(self._reset_unitdata)

    def _reset_unitdata(self):
        unitdata._KV = None


class CharmDirectoryFixture(Fixture):
    def __init__(self, charm_dir='/charm'):
        super().__init__()
        self.charm_dir = charm_dir

    def _setUp(self):
        self.fs = FakeFilesystem()
        if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
            self.__fix_systemfixtures_issues_8()
        self.useFixture(self.fs)
        self.fs.add(self.charm_dir)
        self.fs.add('/srv')
        # setup test charmdir
        shutil.copytree(CHARM_DIR, self.charm_dir)

    def __fix_systemfixtures_issues_8(self):
        """Monkeypatch testing-cabal/systemfixtures/issues/8 fix"""

        def _is_fake_path_or_fd(self, path, *args, **kwargs):
            """Fixed FakeFilesystem._is_fake_path_or_fd"""
            if isinstance(path, int):
                path = self._path_from_fd(path)
            elif isinstance(path, os.DirEntry):
                path = path.name
            return self._is_fake_path(path)

        self.fs._is_fake_path_or_fd = types.MethodType(_is_fake_path_or_fd, self.fs)


class CharmEnvFixture(Fixture):
    def __init__(self, charm_dir='/charm'):
        super().__init__()
        self.charm_dir = charm_dir

    def _setUp(self):
        self.config = self.useFixture(ConfigOverrideFixture())
        self.layer_config = self.useFixture(LayerConfigOverrideFixture())
        self.useFixture(UnitDataFixture())
        self.useFixture(CharmDirectoryFixture(charm_dir=self.charm_dir))
        # set JUJU_CHARM_DIR and CHARM_DIR
        self.useFixture(EnvironmentVariable('JUJU_CHARM_DIR', self.charm_dir))
        self.useFixture(EnvironmentVariable('CHARM_DIR', self.charm_dir))


class ReactiveCharmEnvFixture(Fixture):
    def __init__(self, charm_dir='/charm'):
        super().__init__()
        self.charm_dir = charm_dir

    def _setUp(self):
        self.charm_env = self.useFixture(CharmEnvFixture(charm_dir=self.charm_dir))
        self.config = self.charm_env.config
        self.users = self.useFixture(FakeUsers())
        self.groups = self.useFixture(FakeGroups())
        self.ps = self.useFixture(FakeProcesses())
        log = self.useFixture(MockPatch('charmhelpers.core.hookenv.log'))
        self.log = log.mock

        # override juju-log, as I couldn't find a way to patch hookenv.log
        # when used from hookenv itself. Make it call our mocked log function
        def juju_log(proc_args):
            self.log(proc_args['args'][1])
            return {'returncode': 0}

        self.ps.add(juju_log, name='juju-log')
        # override payload_dir with real CHARM_DIR/files as using overlayed
        # src and dst in shutil.copytree fails. To reproduce change CHARM_DIR
        # with self.charm_dir
        # https://github.com/testing-cabal/systemfixtures/issues/7
        self.useFixture(
            MonkeyPatch(
                'ols.base.payload_dir', lambda: os.path.join(CHARM_DIR, 'files')
            )
        )

        # allow adding user and group
        def addgroup(proc_args):
            self.groups.add(proc_args['args'][-1], 42)
            return {'returncode': 0}

        self.ps.add(addgroup, name='addgroup')

        def useradd(proc_args):
            self.users.add(proc_args['args'][-1], 42)
            return {'returncode': 0}

        self.ps.add(useradd, name='useradd')
        status_set = self.useFixture(MockPatch('charmhelpers.core.hookenv.status_set'))
        self.status_set = status_set.mock
